/**
 * Plot Tooltips
 * @copyright 2022, Firaxis Gmaes
 * @description The tooltips that appear based on the cursor hovering over world plots.
 */

/**
	TCS Improved Plot Tooltip
	-------------------------
	author: thecrazyscotsman
	
	My apologies if my lack of JavaScript or CSS knowledge horrifies you.

 */

console.warn("TCS IMPROVED PLOT TOOLTIP - LOADED");

import TooltipManager, { PlotTooltipPriority } from '/core/ui/tooltips/tooltip-manager.js';
import { ComponentID } from '/core/ui/utilities/utilities-component-id.js';
import DistrictHealthManager from '/base-standard/ui/district/district-health-manager.js';
import LensManager from '/core/ui/lenses/lens-manager.js';
import { TradeRoute } from '/core/ui/utilities/utilities-data.js';
import CityDetails, { UpdateCityDetailsEventName } from "/base-standard/ui/city-details/model-city-details.js";
class PlotTooltipType {
    constructor() {
        this.plotCoord = null;
        this.isShowingDebug = false;
        this.tooltip = document.createElement('fxs-tooltip');
        this.container = document.createElement('div');
        this.yieldsFlexbox = document.createElement('div');
        this.tooltip.classList.add('plot-tooltip', 'max-w-96');
        this.tooltip.appendChild(this.container);
        Loading.runWhenFinished(() => {
            for (const y of GameInfo.Yields) {
                const url = UI.getIcon(`${y.YieldType}`, "YIELD");
                Controls.preloadImage(url, 'plot-tooltip');
            }
        });
    }
    getHTML() {
        return this.tooltip;
    }
    isUpdateNeeded(plotCoord) {
        // Check if the plot location has changed, if not return early, otherwise cache it and rebuild.
        if (this.plotCoord != null) {
            if (plotCoord.x == this.plotCoord.x && plotCoord.y == this.plotCoord.y) {
                return false;
            }
        }
        this.plotCoord = plotCoord; // May be cleaner to recompute in update but at cost of computing 2nd time.
        return true;
    }
    reset() {
        this.container.innerHTML = '';
        this.yieldsFlexbox.innerHTML = '';
    }
    update() {
        if (this.plotCoord == null) {
            console.error("Tooltip was unable to read plot values due to a coordinate error.");
            return;
        }
        this.isShowingDebug = UI.isDebugPlotInfoVisible(); // Ensure debug status hasn't changed
        // Obtain names and IDs
        const plotCoord = this.plotCoord;
        const terrainLabel = this.getTerrainLabel(plotCoord);
        const biomeLabel = this.getBiomeLabel(plotCoord);
        const featureInfo = this.getFeatureLabel(plotCoord); //this is an object, unlike the others
        const continentName = this.getContinentName(plotCoord);
        const riverLabel = this.getRiverLabel(plotCoord);
        const routeName = this.getRouteName();
        const hexResource = this.getResource();
        const playerID = GameplayMap.getOwner(plotCoord.x, plotCoord.y);
        const plotIndex = GameplayMap.getIndexFromLocation(plotCoord);
				
        // Top Section
		// Biome & Terrain
        if (LensManager.getActiveLens() == "fxs-settler-lens") {
            //Add more details to the tooltip if we are in the settler lens
            const localPlayer = Players.get(GameContext.localPlayerID);
            if (!localPlayer) {
                console.error("plot-tooltip: Attempting to update settler tooltip, but no valid local player!");
                return;
            }
            const localPlayerDiplomacy = localPlayer?.Diplomacy;
            if (localPlayerDiplomacy === undefined) {
                console.error("plot-tooltip: Attempting to update settler tooltip, but no valid local player Diplomacy object!");
                return;
            }
            else if (!GameplayMap.isWater(this.plotCoord.x, this.plotCoord.y) && !GameplayMap.isImpassable(this.plotCoord.x, this.plotCoord.y) && !GameplayMap.isNavigableRiver(this.plotCoord.x, this.plotCoord.y)) {
                //Dont't add any extra tooltip to mountains, oceans, or navigable rivers, should be obvious enough w/o them
                const settlerTooltip = document.createElement("div");
                settlerTooltip.classList.add("plot-tooltip__settler-tooltip");
                const localPlayerAdvancedStart = localPlayer?.AdvancedStart;
                if (localPlayerAdvancedStart === undefined) {
                    console.error("plot-tooltip: Attempting to update settler tooltip, but no valid local player advanced start object!");
                    return;
                }
                //Show why we can't settle here
                if (!GameplayMap.isPlotInAdvancedStartRegion(GameContext.localPlayerID, this.plotCoord.x, this.plotCoord.y) && !localPlayerAdvancedStart?.getPlacementComplete()) {
                    settlerTooltip.classList.add("blocked-location");
                    settlerTooltip.innerHTML = Locale.compose("LOC_PLOT_TOOLTIP_CANT_SETTLE_TOO_FAR");
                }
                else if (!localPlayerDiplomacy.isValidLandClaimLocation(this.plotCoord, true /*bIgnoreFriendlyUnitRequirement*/)) {
                    settlerTooltip.classList.add("blocked-location");
                    if (GameplayMap.isCityWithinMinimumDistance(this.plotCoord.x, this.plotCoord.y)) {
                        settlerTooltip.innerHTML = Locale.compose("LOC_PLOT_TOOLTIP_CANT_SETTLE_TOO_CLOSE");
                    }
                    else if (GameplayMap.getResourceType(this.plotCoord.x, this.plotCoord.y) != ResourceTypes.NO_RESOURCE) {
                        settlerTooltip.innerHTML = Locale.compose("LOC_PLOT_TOOLTIP_CANT_SETTLE_RESOURCES");
                    }
                }
                else if (!GameplayMap.isFreshWater(this.plotCoord.x, this.plotCoord.y)) {
                    settlerTooltip.classList.add("okay-location");
                    settlerTooltip.innerHTML = Locale.compose("LOC_PLOT_TOOLTIP_NO_FRESH_WATER");
                }
                this.container.appendChild(settlerTooltip);
                const toolTipHorizontalRule = document.createElement("div");
                toolTipHorizontalRule.classList.add("plot-tooltip__horizontalRule");
                this.container.appendChild(toolTipHorizontalRule);
            }
        }
        const tooltipFirstLine = document.createElement("div");
        tooltipFirstLine.classList.add('text-secondary', 'text-center', 'uppercase', 'font-title');
        if (biomeLabel) {
            // TODO - Add hard-coded string to localization XML.
            const label = Locale.compose("{1_TerrainName} {2_BiomeName}", terrainLabel, biomeLabel);
            tooltipFirstLine.setAttribute('data-l10n-id', label);
        }
        else {
            tooltipFirstLine.setAttribute('data-l10n-id', terrainLabel);
        }
        this.container.appendChild(tooltipFirstLine);
		
		// Feature & River
		if (featureInfo.featureLabel || riverLabel) {
			if (featureInfo.plotIsNaturalWonder) {
				// title line
				this.addTitleLine(featureInfo.featureLabel);
				
				// tooltip line
				this.addHorizontalRule();
				const toolTipNaturalWonderDetails = document.createElement('div');
				toolTipNaturalWonderDetails.classList.add("plot-tooltip__owner-civ-text", "justify-center");
				toolTipNaturalWonderDetails.setAttribute('data-l10n-id', featureInfo.featureTooltip);
				this.container.appendChild(toolTipNaturalWonderDetails);
			}
			else {
				const tooltipSecondLine = document.createElement("div");
				tooltipSecondLine.classList.add('text-2xs', 'text-center');
				if (featureInfo.featureLabel && riverLabel) {
					tooltipSecondLine.setAttribute('data-l10n-id', Locale.compose("LOC_PLOT_MOD_TCS_FEATURE_RIVER", featureInfo.featureLabel, riverLabel));
				}
				else if (featureInfo.featureLabel) {
					tooltipSecondLine.setAttribute('data-l10n-id', featureInfo.featureLabel);
				}
				else {
					tooltipSecondLine.setAttribute('data-l10n-id', riverLabel);
				}
				this.container.appendChild(tooltipSecondLine);
			}
        }
		        
		        
		// Adds info about constructibles, improvements, and wonders to the tooltip
        this.addConstructibleInformation(this.plotCoord);
		
		// Yields Section
        this.yieldsFlexbox.classList.add("plot-tooltip__resourcesFlex");
        this.container.appendChild(this.yieldsFlexbox);
        this.addPlotYields(this.plotCoord, GameContext.localPlayerID);
        if (hexResource) {
			
			// Resource name
			this.addTitleLine(Locale.compose(hexResource.Name));
			
            // Resource icon and tooltip	
			const toolTipResourceContainer = document.createElement('div');
            toolTipResourceContainer.classList.add('plot-tooltip__resource-container');
			
            const toolTipResourceLargeIcon = document.createElement("div");
            toolTipResourceLargeIcon.classList.add("plot-tooltip__large-resource-icon");
			const toolTipResourceIconCSS = UI.getIconCSS(hexResource.ResourceType);
            toolTipResourceLargeIcon.style.backgroundImage = toolTipResourceIconCSS;
            toolTipResourceContainer.appendChild(toolTipResourceLargeIcon);
			
            const toolTipResourceDetails = document.createElement('div');
            toolTipResourceDetails.classList.add('plot-tooltip__resource-details');
			toolTipResourceDetails?.style.setProperty('flex-direction', 'row'); //text needs to flow more smoothly across
			toolTipResourceDetails?.style.setProperty('max-width', '14rem');
			
            const toolTipResourceDescription = document.createElement("div");
            toolTipResourceDescription.classList.add("plot-tooltip__resource-label_description");
            toolTipResourceDescription.setAttribute('data-l10n-id', hexResource.Tooltip);
			
            toolTipResourceDetails.appendChild(toolTipResourceDescription);
            toolTipResourceContainer.appendChild(toolTipResourceDetails);
			
            this.container.appendChild(toolTipResourceContainer);
        }
		
		// Owner Info
		this.addOwnerInfo(this.plotCoord, playerID);
        
		// Unit Info
        this.addUnitInfo(this.plotCoord);
        UI.setPlotLocation(this.plotCoord.x, this.plotCoord.y, plotIndex);
		
		// Continent & Route info
		if (continentName) {
			// Title line
			this.addTitleLine(Locale.compose("LOC_PLOT_MOD_TCS_MORE_INFO"));
			
			// info
			const lastLineLabel = (routeName) ? (continentName + " " + Locale.compose("LOC_PLOT_DIVIDER_DOT") + " " + routeName) : continentName;
			const tooltipLastLine = document.createElement("div");
			tooltipLastLine.classList.add('text-2xs', 'text-center');
			tooltipLastLine.setAttribute('data-l10n-id', lastLineLabel);
			this.container.appendChild(tooltipLastLine);
        }      
		
		// Plot Effects
		this.getPlotEffectNames(plotIndex);
        
		// Adjust cursor between normal and red based on the plot owner's hostility
        if (!UI.isCursorLocked()) {
            const localPlayerID = GameContext.localPlayerID;
            const topUnit = this.getTopUnit(this.plotCoord);
            let showHostileCursor = false;
            let owningPlayerID = GameplayMap.getOwner(this.plotCoord.x, this.plotCoord.y);
            // if there's a unit on the plot, that player overrides the tile's owner
            if (topUnit) {
                owningPlayerID = topUnit.owner;
            }
            const revealedState = GameplayMap.getRevealedState(localPlayerID, plotCoord.x, plotCoord.y);
            if (Players.isValid(localPlayerID) && Players.isValid(owningPlayerID) && (revealedState == RevealedStates.VISIBLE)) {
                const owningPlayer = Players.get(owningPlayerID);
                // Is it an independent?
                if (owningPlayer?.isIndependent) {
                    let independentID = PlayerIds.NO_PLAYER;
                    if (topUnit) {
                        // We got the player from the unit, so use the unit
                        independentID = Game.IndependentPowers.getIndependentPlayerIDFromUnit(topUnit.id);
                    }
                    else {
                        // Get the independent from the plot, can reutrn -1
                        independentID = Game.IndependentPowers.getIndependentPlayerIDAt(this.plotCoord.x, this.plotCoord.y);
                    }
                    if (independentID != PlayerIds.NO_PLAYER) {
                        const relationship = Game.IndependentPowers.getIndependentRelationship(independentID, localPlayerID);
                        if (relationship == IndependentRelationship.HOSTILE) {
                            showHostileCursor = true;
                        }
                    }
                }
                else {
                    var hasHiddenUnit = false;
                    if (topUnit?.hasHiddenVisibility) {
                        hasHiddenUnit = true;
                    }
                    const localPlayer = Players.get(localPlayerID);
                    if (localPlayer) {
                        const localPlayerDiplomacy = localPlayer.Diplomacy;
                        if (localPlayerDiplomacy) {
                            if (localPlayerDiplomacy.isAtWarWith(owningPlayerID) && !hasHiddenUnit) {
                                showHostileCursor = true;
                            }
                        }
                    }
                }
            }
            if (showHostileCursor) {
                UI.setCursorByURL("fs://game/core/ui/cursors/enemy.ani");
            }
            else {
                UI.setCursorByType(UIHTMLCursorTypes.Default);
            }
        }
        
		//debug info
        if (this.isShowingDebug) {
            const tooltipDebugFlexbox = document.createElement("div");
            tooltipDebugFlexbox.classList.add("plot-tooltip__debug-flexbox");
            this.container.appendChild(tooltipDebugFlexbox);
            const toolTipDebugHorizontalRule = document.createElement("div");
            toolTipDebugHorizontalRule.classList.add("plot-tooltip__horizontalRule");
            tooltipDebugFlexbox.appendChild(toolTipDebugHorizontalRule);
            const playerID = GameplayMap.getOwner(this.plotCoord.x, this.plotCoord.y);
            const currHp = Players.Districts.get(playerID)?.getDistrictHealth(this.plotCoord);
            const maxHp = Players.Districts.get(playerID)?.getDistrictMaxHealth(this.plotCoord);
            const toolTipDebugTitle = document.createElement("div");
            toolTipDebugTitle.classList.add("plot-tooltip__debug-title-text");
            if ((currHp != undefined && currHp != 0) && (maxHp != undefined && maxHp != 0)) {
                toolTipDebugTitle.innerHTML = Locale.compose("LOC_PLOT_TOOLTIP_DEBUG_TITLE") + ": " + currHp + " / " + maxHp;
                tooltipDebugFlexbox.appendChild(toolTipDebugTitle);
            }
            else {
                toolTipDebugTitle.innerHTML = Locale.compose("LOC_PLOT_TOOLTIP_DEBUG_TITLE") + ":";
                tooltipDebugFlexbox.appendChild(toolTipDebugTitle);
            }
            const toolTipDebugPlotCoord = document.createElement("div");
            toolTipDebugPlotCoord.classList.add("plot-tooltip__coordinate-text");
            toolTipDebugPlotCoord.innerHTML = Locale.compose("LOC_PLOT_TOOLTIP_PLOT") + `: (${this.plotCoord.x},${this.plotCoord.y})`;
            tooltipDebugFlexbox.appendChild(toolTipDebugPlotCoord);
            const toolTipDebugPlotIndex = document.createElement("div");
            toolTipDebugPlotIndex.classList.add("plot-tooltip__coordinate-text");
            toolTipDebugPlotIndex.innerHTML = Locale.compose("LOC_PLOT_TOOLTIP_INDEX") + `: ${plotIndex}`;
            tooltipDebugFlexbox.appendChild(toolTipDebugPlotIndex);
            const localPlayer = Players.get(GameContext.localPlayerID);
            if (localPlayer != null) {
                if (localPlayer.isDistantLands(this.plotCoord)) {
                    const toolTipDebugPlotTag = document.createElement("div");
                    toolTipDebugPlotTag.classList.add("plot-tooltip__coordinate-text");
                    toolTipDebugPlotTag.innerHTML = Locale.compose("LOC_PLOT_TOOLTIP_HEMISPHERE_WEST");
                    tooltipDebugFlexbox.appendChild(toolTipDebugPlotTag);
                }
                else {
                    const toolTipDebugPlotTag = document.createElement("div");
                    toolTipDebugPlotTag.classList.add("plot-tooltip__coordinate-text");
                    toolTipDebugPlotTag.innerHTML = Locale.compose("LOC_PLOT_TOOLTIP_HEMISPHERE_EAST");
                    tooltipDebugFlexbox.appendChild(toolTipDebugPlotTag);
                }
            }
        }
    }
    isBlank() {
        if (this.plotCoord == null) {
            return true;
        }
        const localPlayerID = GameContext.localPlayerID;
        const revealedState = GameplayMap.getRevealedState(localPlayerID, this.plotCoord.x, this.plotCoord.y);
        if (revealedState == RevealedStates.HIDDEN) {
            return true;
        }
        // If a unit is selected, check if over our own unit an enemy unit and prevent the plot tooltip from displaying.
        const selectedUnitID = UI.Player.getHeadSelectedUnit();
        if (selectedUnitID && ComponentID.isValid(selectedUnitID)) {
            const plotUnits = MapUnits.getUnits(this.plotCoord.x, this.plotCoord.y);
            if (plotUnits.length > 0) {
                // Hovering over your selected unit; don't show the plot tooltip
                if (plotUnits.find(e => ComponentID.isMatch(e, selectedUnitID))) {
                    return true;
                }
                let args = {};
                args.X = this.plotCoord.x;
                args.Y = this.plotCoord.y;
                let combatType = Game.Combat.testAttackInto(selectedUnitID, args);
                if (combatType != CombatTypes.NO_COMBAT) {
                    return true;
                }
            }
        }
        return false;
    }
    
	addHorizontalRule() {
		const toolTipHorizontalRule = document.createElement("div");
		toolTipHorizontalRule.classList.add("plot-tooltip__horizontalRule");
		this.container.appendChild(toolTipHorizontalRule);
	}
	addSeparatorLine() {
		const tooltipSeparator_LineFlex = document.createElement("div");
		tooltipSeparator_LineFlex.classList.add("plot-tooltip__TitleLineFlex");
		const titleLeftSeparator = document.createElement("div");
		titleLeftSeparator.classList.add("plot-tooltip__TitleLineleft");
		tooltipSeparator_LineFlex.appendChild(titleLeftSeparator);
		const titleRightSeparator = document.createElement("div");
		titleRightSeparator.classList.add("plot-tooltip__TitleLineRight");
		tooltipSeparator_LineFlex.appendChild(titleRightSeparator);
		this.container.appendChild(tooltipSeparator_LineFlex);
	}
	addGradiantBackground(linearGradiantLeft, linearGradiantRight, container) {
		const tooltipSeparator_LineFlex = document.createElement("div");
		tooltipSeparator_LineFlex.classList.add("plot-tooltip__TitleLineFlex");
		const titleLeftSeparator = document.createElement("div");
		titleLeftSeparator.classList.add("plot-tooltip__TitleLineleft");
		titleLeftSeparator?.style.setProperty("height", '100%');
		titleLeftSeparator?.style.setProperty("width", '50%');
		titleLeftSeparator?.style.setProperty('background-image', linearGradiantLeft);
		titleLeftSeparator?.style.setProperty('align-self', 'flex-start');
		tooltipSeparator_LineFlex.appendChild(titleLeftSeparator);
		const titleRightSeparator = document.createElement("div");
		titleRightSeparator.classList.add("plot-tooltip__TitleLineRight");
		titleRightSeparator?.style.setProperty("height", '100%');
		titleRightSeparator?.style.setProperty("width", '50%');
		titleRightSeparator?.style.setProperty('background-image', linearGradiantRight);
		titleRightSeparator?.style.setProperty('align-self', 'flex-start');
		tooltipSeparator_LineFlex.appendChild(titleRightSeparator);
		container.appendChild(tooltipSeparator_LineFlex);
	}
	addTitleLine(titleText) {
		const tooltipTitle_LineFlex = document.createElement("div");
		tooltipTitle_LineFlex.classList.add("plot-tooltip__TitleLineFlex");
		const titleLeftSeparator = document.createElement("div");
		titleLeftSeparator.classList.add("plot-tooltip__TitleLineleft");
		titleLeftSeparator?.style.setProperty('min-width', '0.5rem');
		tooltipTitle_LineFlex.appendChild(titleLeftSeparator);
		const tooltipTitleName = document.createElement("div");
		tooltipTitleName.classList.add("plot-tooltip__ImprovementName");
		tooltipTitleName?.style.setProperty('max-width', '12rem');
		tooltipTitleName.innerHTML = titleText;
		tooltipTitle_LineFlex.appendChild(tooltipTitleName);
		const titleRightSeparator = document.createElement("div");
		titleRightSeparator.classList.add("plot-tooltip__TitleLineRight");
		titleRightSeparator?.style.setProperty('min-width', '0.5rem');
		tooltipTitle_LineFlex.appendChild(titleRightSeparator);
		this.container.appendChild(tooltipTitle_LineFlex);
	}
	
	getContinentName(location) {
        const continentType = GameplayMap.getContinentType(location.x, location.y);
        const continent = GameInfo.Continents.lookup(continentType);
        if (continent && continent.Description) {
            return Locale.compose(continent.Description);
        }
        else {
            return "";
        }
    }
    addConstructibleInformation(plotCoordinate) {
		
		/*
		Format:
		
		district (rural/urban/quarter/wonder)
		constructibles (improvements/buildings)
		*/
		
		// Get constructibles
		const buildings = [];
		const walls = [];
        const wonders = [];
        const improvements = [];
        const constructibles = MapConstructibles.getHiddenFilteredConstructibles(plotCoordinate.x, plotCoordinate.y);
		
        constructibles.forEach((item) => {
            const instance = Constructibles.getByComponentID(item);
			if (!instance) {
				return;
			}
			const location = instance.location;
			if (location.x != plotCoordinate.x || location.y != plotCoordinate.y) {
				return;
			}
			const info = GameInfo.Constructibles.lookup(instance.type);
			if (!info) {
				console.warn("Building constructible without a definition: " + instance.type.toString());
				return;
			}
			if (info.ConstructibleClass == "BUILDING") {
				const constructibleType = info.ConstructibleType;
				const building = {
					Info: info,
					Name: Locale.compose(info.Name), //LOC_PLOT_MOD_TCS_CONSTRUCTIBLE_NAME -> name + tag
					Ageless: GameInfo.TypeTags.find(e => e.Tag == "AGELESS" && e.Type == info.ConstructibleType), //LOC_PLOT_MOD_TCS_AGELESS
					UniqueTrait: GameInfo.Buildings.find(e => e.ConstructibleType == info.ConstructibleType && e.TraitType !== null), //LOC_PLOT_MOD_TCS_UNIQUE
					ConstructibleAge: Database.makeHash(info?.Age ?? ""), //LOC_PLOT_MOD_TCS_OBSOLETE
					Damaged: instance.damaged, //LOC_PLOT_TOOLTIP_DAMAGED
					Completed: (instance.complete) //LOC_PLOT_TOOLTIP_IN_PROGRESS
				}
				if (info.DistrictDefense > 0 && info.ExistingDistrictOnly > 0) {
					walls.push(building);
				}
				else {
					buildings.push(building);
				}
			}
			else if (info.ConstructibleClass == "WONDER") {
				const wonder = {
					Info: info,
					Name: Locale.compose(info.Name), //LOC_PLOT_MOD_TCS_CONSTRUCTIBLE_NAME -> name + tag
					Tooltip : Locale.compose(info.Tooltip),
					ConstructibleAge: Database.makeHash(info?.Age ?? ""), //LOC_PLOT_MOD_TCS_OBSOLETE
					Damaged: instance.damaged, //LOC_PLOT_TOOLTIP_DAMAGED
					Completed: (instance.complete) //LOC_PLOT_TOOLTIP_IN_PROGRESS
				}
				wonders.push(wonder);
			}
			else if (info.ConstructibleClass == "IMPROVEMENT") {
				const improvement = {
					Info: info,
					Name: Locale.compose(info.Name), //LOC_PLOT_MOD_TCS_CONSTRUCTIBLE_NAME -> name + tag
					//Ageless: GameInfo.TypeTags.find(e => e.Tag == "AGELESS" && e.Type == info.ConstructibleType), //LOC_PLOT_MOD_TCS_AGELESS
					UniqueTrait: GameInfo.Improvements.find(e => e.ConstructibleType == info.ConstructibleType && e.TraitType !== null), //LOC_PLOT_MOD_TCS_UNIQUE
					//ConstructibleAge: Database.makeHash(info?.Age ?? ""), //LOC_PLOT_MOD_TCS_OBSOLETE
					Damaged: instance.damaged, //LOC_PLOT_TOOLTIP_DAMAGED
					Completed: (instance.complete) //LOC_PLOT_TOOLTIP_IN_PROGRESS
				}
				improvements.push(improvement);
			}
			else {
				const pass = true;
			}
        });
		
		const iCurrentAge = Game.age;
		const districtId = MapCities.getDistrict(plotCoordinate.x, plotCoordinate.y);
		const district = (districtId) ? Districts.get(districtId) : undefined;
		let districtName = (district) ? Locale.compose(GameInfo.Districts.lookup(district.type).Name) : Locale.compose("LOC_PLOT_MOD_TCS_WILDERNESS");
		
		// Check for Quarters
		const quarters = [];
		if (district && district.type == DistrictTypes.URBAN && buildings.length > 1) {
			const uniques = [];
			const ages = [];
			buildings.forEach((item) => {
				if (item.UniqueTrait) {
					uniques.push(item.UniqueTrait.TraitType);
				}
				if (item.ConstructibleAge) {
					ages.push(item.ConstructibleAge);
				}
			});
			if (uniques.length > 1) {
				const uniquesSet = new Set(uniques);
				if (uniquesSet.size == 1) {
					const uniqueQuarter = GameInfo.UniqueQuarters.find(e => e.TraitType == uniques[0]);
					const civType = GameInfo.LegacyCivilizationTraits.find(e => e.TraitType == uniques[0]);
					const civLegacy = GameInfo.LegacyCivilizations.find(e => e.CivilizationType == civType.CivilizationType);
					const civAdjective = Locale.compose(civLegacy.Adjective);
					const quarter = {
						QuarterName: uniqueQuarter.Name,
						QuarterDescription: uniqueQuarter.Description,
						QuarterTooltip: "LOC_" + uniqueQuarter.UniqueQuarterType + "_TOOLTIP",
						Civilization: civLegacy,
						CivilizationAdjective: civLegacy.Adjective
					}
					quarters.push(quarter);
				}
			}
			else if (ages.length > 1) {
				const agesSet = new Set(ages);
				if (agesSet.size == 1 && ages[0] == iCurrentAge) {
					const quarter = {
						QuarterName: "LOC_PLOT_MOD_TCS_QUARTER",
						QuarterDescription: undefined,
						Civilization: undefined,
						CivilizationAdjective: undefined
					}
					quarters.push(quarter);
				}
			}
			else {
				const pass = true;
			}
		}
		
		// Fortifications Container
		const playerID = GameplayMap.getOwner(plotCoordinate.x, plotCoordinate.y);
		const player = Players.get(playerID);
		if (player) {
			if (district && district.type != DistrictTypes.RURAL) {
				const playerDistricts = Players.Districts.get(playerID);
				if (playerDistricts) {
					const currentHealth = playerDistricts.getDistrictHealth(plotCoordinate);
					const maxHealth = playerDistricts.getDistrictMaxHealth(plotCoordinate);
					const isFortified = (maxHealth > 0 && currentHealth == maxHealth) ? true : false;
					const isDamaged = (maxHealth > 0 && currentHealth < maxHealth) ? true : false;
					const isUnderSiege = playerDistricts.getDistrictIsBesieged(plotCoordinate);
					const isHealing = (!isUnderSiege && isDamaged) ? true : false;
					
					//Wall info
					let wallName;
					if (walls.length > 0) {
						for (const wall of walls) {
							if (wall.Info.ConstructibleType == "BUILDING_DEFENSIVE_FORTIFICATIONS" ) {
								wallName = wall.Name;
								break;
							}
							else if (wall.Info.ConstructibleType == "BUILDING_MEDIEVAL_WALLS" ) {
								wallName = wall.Name;
								break;
							}
							else if (wall.Info.ConstructibleType == "BUILDING_ANCIENT_WALLS" ) {
								wallName = wall.Name;
								break;
							}
							else {
								wallName = wall.Name;
							}
						}
					}
					if (isFortified && !isDamaged) {
						this.addHorizontalRule();
						//const districtContainer = document.createElement("div");
						//districtContainer.classList.add("plot-tooltip__district-container");
						//districtContainer?.style.setProperty("background-color", "transparent");
						//this.addGradiantBackground("linear-gradient(to left, #507552, rgba(80, 117, 82, 0))", "linear-gradient(to right, #507552, rgba(80, 117, 82, 0))", districtContainer);
						const districtTitle = document.createElement("div");
						districtTitle.classList.add("text-2xs", "text-center");
						districtTitle.innerHTML = (!wallName) ? Locale.stylize("LOC_PLOT_MOD_TCS_FORTIFIED", currentHealth, maxHealth) : Locale.stylize("LOC_PLOT_MOD_TCS_FORTIFIED_WALL", wallName, currentHealth, maxHealth);
						/*const districtHealth = document.createElement("div");
						districtHealth.classList.add("plot-tooltip__district-health");
						const healthCaption = document.createElement("div");
						healthCaption.classList.add("plot-tooltip__lineThree");
						healthCaption?.style.removeProperty("color"); //remove text color change
						healthCaption?.style.setProperty("font-size", "calc(1rem + -0.3333333333rem)");
						healthCaption.innerHTML = (!wallName) ? Locale.compose("LOC_PLOT_MOD_TCS_DISTRICT_HEALTH", currentHealth, maxHealth) : wallName + " " + Locale.compose("LOC_PLOT_DIVIDER_DOT") + " " + Locale.compose("LOC_PLOT_MOD_TCS_DISTRICT_HEALTH", currentHealth, maxHealth);
						districtHealth.appendChild(healthCaption);
						districtContainer.appendChild(districtTitle);
						districtContainer.appendChild(districtHealth);*/
						this.container.appendChild(districtTitle);
					}
					if (isUnderSiege || isHealing) {
						this.addHorizontalRule();
						const districtContainer = document.createElement("div");
						districtContainer.classList.add("plot-tooltip__district-container");
						//districtContainer?.style.removeProperty("background-color");
						//this.addGradiantBackground("linear-gradient(to left, #3a0806, rgba(58, 8, 6, 0))", "linear-gradient(to right, #3a0806, rgba(58, 8, 6, 0))", districtContainer);
						const districtTitle = document.createElement("div");
						districtTitle.classList.add("plot-tooltip__district-title", "plot-tooltip__lineThree");
						districtTitle.innerHTML = isUnderSiege ? Locale.compose("LOC_PLOT_TOOLTIP_UNDER_SIEGE") : Locale.compose("LOC_PLOT_TOOLTIP_HEALING_DISTRICT");
						const districtHealth = document.createElement("div");
						districtHealth.classList.add("plot-tooltip__district-health");
						const healthCaption = document.createElement("div");
						healthCaption.classList.add("plot-tooltip__health-caption", "plot-tooltip__lineThree");
						healthCaption.innerHTML = (!wallName) ? Locale.compose("LOC_PLOT_MOD_TCS_DISTRICT_HEALTH", currentHealth, maxHealth) : wallName + " " + Locale.compose("LOC_PLOT_DIVIDER_DOT") + " " + Locale.compose("LOC_PLOT_MOD_TCS_DISTRICT_HEALTH", currentHealth, maxHealth);
						districtHealth.appendChild(healthCaption);
						districtContainer.appendChild(districtTitle);
						districtContainer.appendChild(districtHealth);
						this.container.appendChild(districtContainer);
					}	
				}
			}
		}
		
		//Title Line
		if (player || (improvements.length > 0)) {
			if (district && district.type == DistrictTypes.WONDER) {
				districtName = wonders[0].Name;
				this.addTitleLine(districtName);
			}
			else if (quarters.length > 0) {
				districtName = Locale.compose(quarters[0].QuarterName);
				this.addTitleLine(districtName);
				if (quarters[0].QuarterDescription) {
					// Add description
					this.addHorizontalRule();
					const toolTipQuarterDetails = document.createElement('div');
					//toolTipWonderDetails.classList.add("plot-tooltip__resource-label_description");
					toolTipQuarterDetails.classList.add("plot-tooltip__owner-civ-text", "justify-center");
					toolTipQuarterDetails.setAttribute('data-l10n-id', Locale.compose("LOC_PLOT_MOD_TCS_UNIQUE_QUARTER", quarters[0].CivilizationAdjective) + " " + Locale.compose(quarters[0].QuarterTooltip));
					this.container.appendChild(toolTipQuarterDetails);
				}
			}
			else {
				this.addTitleLine(districtName);
			}
		}
			
		// Improvements
        if (improvements.length > 0) {
			improvements.sort((a,b) => (a.Name > b.Name) ? 1 : ((b.Name > a.Name) ? -1 : 0));
            improvements.forEach((item) => {
				
				let itemName = item.Name;
				if (item.UniqueTrait) {
					itemName = item.Name + " " + Locale.compose("LOC_PLOT_MOD_TCS_UNIQUE");
				}
				
				let itemStatus = '';
				if (item.Damaged) {
					itemStatus = "(" + Locale.compose("LOC_PLOT_TOOLTIP_DAMAGED") + ")";
				}
				
                const tooltipImprovementLineFlex = document.createElement("div");
                tooltipImprovementLineFlex.classList.add("plot-tooltip__horizontalRule");
                this.container.appendChild(tooltipImprovementLineFlex);
                const tooltipImprovementName = document.createElement("div");
                tooltipImprovementName.classList.add("plot-tooltip__owner-leader-text");
                tooltipImprovementName.innerHTML = Locale.compose("LOC_PLOT_MOD_TCS_CONSTRUCTIBLE_STATUS", itemName, itemStatus);
                tooltipImprovementLineFlex.appendChild(tooltipImprovementName);
            });
        }
		
		// Buildings
        if (buildings.length > 0) {
			buildings.sort((a,b) => (a.Name > b.Name) ? 1 : ((b.Name > a.Name) ? -1 : 0));
            buildings.forEach((item) => {
				let itemName;
				if (item.UniqueTrait) {
					itemName = item.Name + " " + Locale.compose("LOC_PLOT_MOD_TCS_UNIQUE");
				}
				else if (item.Ageless) {
					itemName = item.Name + " " + Locale.compose("LOC_PLOT_MOD_TCS_AGELESS");
				}
				else if (item.ConstructibleAge != iCurrentAge) {
					itemName = item.Name + " " + Locale.compose("LOC_PLOT_MOD_TCS_OBSOLETE");
				}
				else {
					itemName = item.Name;
				}
				
				let itemStatus = '';
				if (item.Damaged) {
					itemStatus = "(" + Locale.compose("LOC_PLOT_TOOLTIP_DAMAGED") + ")";
				}
				else if (!item.Completed) {
					itemStatus = "(" + Locale.compose("LOC_PLOT_TOOLTIP_IN_PROGRESS") + ")";
				}
				else {
					const pass = true;
				}
				
                const tooltipImprovementLineFlex = document.createElement("div");
                tooltipImprovementLineFlex.classList.add("plot-tooltip__horizontalRule");
                this.container.appendChild(tooltipImprovementLineFlex);
                const tooltipImprovementName = document.createElement("div");
                tooltipImprovementName.classList.add("plot-tooltip__owner-leader-text");
                tooltipImprovementName.innerHTML = Locale.compose("LOC_PLOT_MOD_TCS_CONSTRUCTIBLE_STATUS", itemName, itemStatus);
                tooltipImprovementLineFlex.appendChild(tooltipImprovementName);
            });
        }
		
		// Wonders
        if (wonders.length > 0) {
			wonders.sort((a,b) => (a.Name > b.Name) ? 1 : ((b.Name > a.Name) ? -1 : 0));		
            wonders.forEach((item) => {         
				let itemStatus;
				if (item.Damaged) {
					itemStatus = "(" + Locale.compose("LOC_PLOT_TOOLTIP_DAMAGED") + ")";
				}
				else if (!item.Completed) {
					itemStatus = "(" + Locale.compose("LOC_PLOT_TOOLTIP_IN_PROGRESS") + ")";
				}
				else {
					const pass = true;
				}
				// Add status
				if (itemStatus) {
					const tooltipBuildingStatus = document.createElement("div");
					tooltipBuildingStatus.classList.add("plot-tooltip__building-status");
					tooltipBuildingStatus.innerHTML = itemStatus;
					this.container.appendChild(tooltipBuildingStatus);
				}			
				
				// Add description
				this.addHorizontalRule();
				const toolTipWonderDetails = document.createElement('div');
				//toolTipWonderDetails.classList.add("plot-tooltip__resource-label_description");
				toolTipWonderDetails.classList.add("plot-tooltip__owner-civ-text", "justify-center");
				toolTipWonderDetails.setAttribute('data-l10n-id', item.Tooltip);
				this.container.appendChild(toolTipWonderDetails);
            });
        }
    }
    getPlayerName() {
        const playerID = GameplayMap.getOwner(this.plotCoord.x, this.plotCoord.y);
        const player = Players.get(playerID);
        if (player == null) {
            return "";
        }
        const localPlayerID = GameContext.localPlayerID;
        const name = Locale.stylize(player.name) + ((playerID == localPlayerID) ? (" (" + Locale.compose("LOC_PLOT_TOOLTIP_YOU") + ")") : "");
        return name;
    }
    getCivName() {
        const playerID = GameplayMap.getOwner(this.plotCoord.x, this.plotCoord.y);
        const player = Players.get(playerID);
        if (player == null) {
            return "";
        }
        const name = Locale.compose(GameplayMap.getOwnerName(this.plotCoord.x, this.plotCoord.y));
        return name;
    }
    getTerrainLabel(location) {
        const terrainType = GameplayMap.getTerrainType(location.x, location.y);
        const terrain = GameInfo.Terrains.lookup(terrainType);
        if (terrain) {
            if (this.isShowingDebug) {
                // despite being "coast" this is a check for a lake
                if (terrain.TerrainType == "TERRAIN_COAST" && GameplayMap.isLake(location.x, location.y)) {
                    return Locale.compose('{1_Name} ({2_Value})', "LOC_TERRAIN_LAKE_NAME", terrainType.toString());
                }
                return Locale.compose('{1_Name} ({2_Value})', terrain.Name, terrainType.toString());
            }
            else {
                // despite being "coast" this is a check for a lake
                if (terrain.TerrainType == "TERRAIN_COAST" && GameplayMap.isLake(location.x, location.y)) {
                    return "LOC_TERRAIN_LAKE_NAME";
                }
                return terrain.Name;
            }
        }
        else {
            return "";
        }
    }
    getBiomeLabel(location) {
        const biomeType = GameplayMap.getBiomeType(location.x, location.y);
        const biome = GameInfo.Biomes.lookup(biomeType);
        // Do not show a label if marine biome.
        if (biome && biome.BiomeType != "BIOME_MARINE") {
            if (this.isShowingDebug) {
                return Locale.compose('{1_Name} ({2_Value})', biome.Name, biomeType.toString());
            }
            else {
                return biome.Name;
            }
        }
        else {
            return "";
        }
    }
    getResource() {
        if (this.plotCoord) {
            const resourceType = GameplayMap.getResourceType(this.plotCoord.x, this.plotCoord.y);
            return GameInfo.Resources.lookup(resourceType);
        }
        return null;
    }
    getRouteName() {
        const routeType = GameplayMap.getRouteType(this.plotCoord.x, this.plotCoord.y);
        const route = GameInfo.Routes.lookup(routeType);
        const isFerry = GameplayMap.isFerry(this.plotCoord.x, this.plotCoord.y);
        let returnString = "";
        if (route) {
            if (isFerry) {
                returnString = Locale.compose(route.Name) + " " + Locale.compose("LOC_PLOT_DIVIDER_DOT") + " " + Locale.compose("LOC_NAVIGABLE_RIVER_FERRY");
            }
            else {
                returnString = Locale.compose(route.Name);
            }
        }
        return returnString;
    }
    getPlotEffectNames(plotIndex) {
        const plotEffects = MapPlotEffects.getPlotEffects(plotIndex);
		if (plotEffects) {
			const localPlayerID = GameContext.localPlayerID;
			const plotEffectsFiltered = plotEffects.filter(item => (!item.onlyVisibleToOwner || (item.onlyVisibleToOwner && (item.owner == localPlayerID))));
			if (plotEffectsFiltered.length > 0) {
				let effectString;
				const toolTipPlotEffectsHorizontalRule = document.createElement("div");
				toolTipPlotEffectsHorizontalRule.classList.add("plot-tooltip__horizontalRule");
				this.container.appendChild(toolTipPlotEffectsHorizontalRule);
				const toolTipPlotEffectsText = document.createElement("div");
				toolTipPlotEffectsText.classList.add('text-2xs', 'text-center');
				plotEffectsFiltered?.forEach((item) => {
					const effectInfo = GameInfo.PlotEffects.lookup(item.effectType);
					if (effectInfo) {
						if (!effectString) {
							effectString = Locale.compose(effectInfo.Name);
						}
						else {
							effectString = effectString + " " + Locale.compose("LOC_PLOT_DIVIDER_DOT") + " " + Locale.compose(effectInfo.Name);
						}					
					}
				});
				toolTipPlotEffectsText.setAttribute('data-l10n-id', effectString);
				this.container.appendChild(toolTipPlotEffectsText);
			}
		}
    }
    getTopUnit(location) {
        let plotUnits = MapUnits.getUnits(location.x, location.y);
        if (plotUnits && plotUnits.length > 0) {
            const topUnit = Units.get(plotUnits[0]);
            return topUnit;
        }
        return null;
    }
    /**
     * Add to a plot tooltip information on the plot's owner
     * @param {float2} location The X,Y plot location.
     * @param {playerId} playerID The player associated with the request.
     */
    addOwnerInfo(location, owningPlayerID) {
		
		// Cancel if player is not alive
		const owningPlayer = Players.get(owningPlayerID);
		if (!owningPlayer || !Players.isAlive(owningPlayerID)) {
            return;
        }
        const filteredConstructibles = MapConstructibles.getHiddenFilteredConstructibles(location.x, location.y);
        const constructibles = MapConstructibles.getConstructibles(location.x, location.y);        
        if (filteredConstructibles.length == 0 && filteredConstructibles.length != constructibles.length) {
            return;
        }
		
		const owningPlayerDiplomacy = (owningPlayer) ? owningPlayer.Diplomacy : undefined;
		
		// Get local player
		const localPlayerID = GameContext.localPlayerID;
		const localPlayer = Players.get(localPlayerID);
		const localPlayerDiplomacy = localPlayer?.Diplomacy;  
		
		// Get ownership labels (playerLabel, civLabel, cityLabel)
		const playerLabel = (owningPlayer) ? Locale.stylize(owningPlayer.name) + ((owningPlayerID == localPlayerID) ? (" (" + Locale.compose("LOC_PLOT_TOOLTIP_YOU") + ")") : "") : undefined;
		const civLabel = Locale.compose(GameplayMap.getOwnerName(location.x, location.y));
		let cityLabel;
		let cityHappiness;
		let townFocus;
		let tradeRoutes;
		let connectedSettlementString = "";
		const connectedSettlementNames = [];
		const citiesReceivingFood = [];
		if (owningPlayer && Players.isAlive(owningPlayerID)) {
			const playerCities = owningPlayer.Cities;
			if (playerCities) {
				const cities = playerCities.getCities();
				for (const city of cities) {
					if (ComponentID.isMatch(city.id, GameplayMap.getOwningCityFromXY(location.x, location.y)) == true) {
						cityLabel = Locale.compose("LOC_PLOT_MOD_TCS_CITY", Locale.compose(city.name));
						if (city.isTown) {
							cityLabel = Locale.stylize("LOC_PLOT_MOD_TCS_TOWN", Locale.compose(city.name));
						}
						else {
							cityLabel = Locale.stylize("LOC_PLOT_MOD_TCS_CITY", Locale.compose(city.name));
						}
						
						// Get connections
						const connectedSettlements = city.getConnectedCities();
						if (connectedSettlements.length > 0) {
							// Get town food exports
							if (city.isTown) {
								const townFoodYield = city.Yields?.getNetYield(YieldTypes.YIELD_FOOD);
								if (townFoodYield) {
									for (const townConnectedSettlementID of connectedSettlements) {
										const townConnectedSettlement = Cities.get(townConnectedSettlementID);
										if (townConnectedSettlement && !townConnectedSettlement.isTown) {
											citiesReceivingFood.push(townConnectedSettlement);
										}
									}	
								}
							}
							// Build connections string	
							for (const connectedsettlementID of connectedSettlements) {
								const connectedSettlement = Cities.get(connectedsettlementID);
								if (connectedSettlement.isTown) {
									connectedSettlementNames.push(Locale.compose("LOC_PLOT_MOD_TCS_TOWN", connectedSettlement.name));
								}
								else {
									connectedSettlementNames.push(Locale.compose("LOC_PLOT_MOD_TCS_CITY", connectedSettlement.name));
								}
							}
							connectedSettlementNames.sort();
							/*for (const connectedsettlementName of connectedSettlementNames) {
								if (connectedSettlementString == "") {
									connectedSettlementString = connectedsettlementName;
								}
								else {
									connectedSettlementString = connectedSettlementString + "[N]" + connectedsettlementName;
								}
							}*/							
						}
						
						// Get Town Focus
						if (city.isTown) {
							if (city.Growth?.growthType == GrowthTypes.EXPAND) {
								townFocus = Locale.compose("LOC_PLOT_MOD_TCS_TOWN_GROWING");
							}
							else {
								townFocus = Locale.compose("LOC_PLOT_MOD_TCS_TOWN_SPECIALIZED");		
							}
							
							//There are apparantly only 2 GrowthTypes: GrowthTypes.EXPAND and GrowthTypes.PROJECT.
							//EXPAND indicates a growing town, PROJECT indicates a specialized town. Each specialization has an entry in the Projects table.
							
							//TO DO: investigate further to see if we can figure out what Project a town has.
							//Potential info in production-chooser-helpers.js
						}
						
						/*
							Additional information which can be gained from the city object:
								Population::
									city.population
								Turns until growth::
									city.Growth.turnsUntilGrowth
								Local Happiness::
									city.Yields?.getYield(YieldTypes.YIELD_HAPPINESS)
								Original Owner (a player id):
									city.originalOwner
								Current Owner (a player id):
									city.owner	
						*/
						
						// Get Happiness
						/*const happiness = city.Yields?.getYield(YieldTypes.YIELD_HAPPINESS);
						if (happiness) {
							cityHappiness = Locale.compose("LOC_PLOT_MOD_TCS_HAPPY");
							if (happiness < 0) {
								cityHappiness = Locale.compose("LOC_PLOT_MOD_TCS_UNHAPPY");
							}
							else if (happiness < -10) {
								cityHappiness = Locale.compose("LOC_PLOT_MOD_TCS_ANGRY");
							}
							if (city.isInfected) {
								cityHappiness = Locale.compose("LOC_PLOT_MOD_TCS_PLAGUE");
							}
						}*/
						
						// Get Trade Routes
						/*const routes = city.Trade.routes;
						if (routes.length > 0) {
							for (let route of routes) {
								// Resources from this route? Include partner city and list resource yields
								const myRoutePayload = TradeRoute.getCityPayload(route, city.id);
								if (myRoutePayload && myRoutePayload.resourceValues.length > 0) {
									// Partner city name
									let partnerCity = TradeRoute.getOppositeCity(route, city.id);
									if (partnerCity) {
										tradeRoutes = (tradeRoutes) ? (tradeRoutes + " " + Locale.compose("LOC_PLOT_DIVIDER_DOT") + " " + Locale.compose(partnerCity.name)) : Locale.compose(partnerCity.name);
									}
								}
							}
						}*/
						break;
					}
				}
			}
		}
		
		// Get relationship label (relationshipLabel)
		let relationshipLabel;
		if (owningPlayer && owningPlayerID && owningPlayerID != localPlayerID) {
			// If Plot Owner is an Independent and the local player is its suzerain, display relationship
			if (owningPlayer.isMinor && owningPlayer.Influence?.hasSuzerain && owningPlayer.Influence.getSuzerain() == localPlayerID) {
				relationshipLabel = Locale.compose("LOC_PLOT_MOD_TCS_RELATIONSHIP_VASSAL")
			}
			else {
				// Get diplomacy
				
				// TO DO: see if I can get active Open Borders deals.
				if (owningPlayerDiplomacy.isAtWarWith(GameContext.localPlayerID)) {
					relationshipLabel = Locale.compose("LOC_PLOT_MOD_TCS_RELATIONSHIP_WAR");
				}
				else if (owningPlayerDiplomacy.hasAllied(localPlayerID)) {
					relationshipLabel = Locale.compose("LOC_PLOT_MOD_TCS_RELATIONSHIP_ALLIANCE", Locale.compose(owningPlayerDiplomacy.getRelationshipLevelName(GameContext.localPlayerID)));
				}
				else {
					relationshipLabel = Locale.compose("LOC_PLOT_MOD_TCS_RELATIONSHIP_OTHER", Locale.compose(owningPlayerDiplomacy.getRelationshipLevelName(GameContext.localPlayerID)));
				}
			}
		}
		
		// Get district info
		const districtId = MapCities.getDistrict(location.x, location.y);
		const district = (districtId) ? Districts.get(districtId) : undefined;
		const plotIsCityCenter = (district && district.type == DistrictTypes.CITY_CENTER);
		
		//================
		// Build Container
		//================
		
		// Leader & Civ & Relationship, City & Status
		if (playerLabel) {
			// City
			if (cityLabel) {
				this.addTitleLine(cityLabel);
			}
			else {
				this.addTitleLine(civLabel);
			}
			
			// Player & Civ
			if (civLabel) {
				this.addHorizontalRule();
				const plotTooltipOwnerLeader = document.createElement("div");
				plotTooltipOwnerLeader.classList.add("plot-tooltip__owner-leader-text");
				if (owningPlayer.isIndependent) {
					plotTooltipOwnerLeader.innerHTML = (relationshipLabel) ? (playerLabel + " " + relationshipLabel) : playerLabel;
				}
				else {
					plotTooltipOwnerLeader.innerHTML = (relationshipLabel) ? (playerLabel + " " + relationshipLabel + " " + Locale.compose("LOC_PLOT_DIVIDER_DOT") + " " + civLabel) : (playerLabel + " " + Locale.compose("LOC_PLOT_DIVIDER_DOT") + " " + civLabel);
				}
				this.container.appendChild(plotTooltipOwnerLeader);
			}
			
			// Additional info for local player when hovering over city center
			if (localPlayerID == owningPlayerID) {
				// Town Focus
				if (townFocus) {
					this.addHorizontalRule();
					const plotTooltipTownFocus = document.createElement("div");
					plotTooltipTownFocus.classList.add('text-2xs', 'text-center');
					let tooltipTownFocusString;
					if (plotIsCityCenter) {
						tooltipTownFocusString = (citiesReceivingFood.length == 1) ? (townFocus + " " + Locale.compose("LOC_PLOT_DIVIDER_DOT") + " " + Locale.compose("LOC_PLOT_MOD_TCS_TOWN_FEEDING_SINGULAR")) : (Locale.compose("LOC_PLOT_MOD_TCS_TOWN_SPECIALIZED") + " " + Locale.compose("LOC_PLOT_DIVIDER_DOT") + " " + Locale.compose("LOC_PLOT_MOD_TCS_TOWN_FEEDING_PLURAL", citiesReceivingFood.length));
					}
					else {
						tooltipTownFocusString = townFocus;
					}
					plotTooltipTownFocus.innerHTML = tooltipTownFocusString;
					this.container.appendChild(plotTooltipTownFocus);
				}
				
				// Connections
				if (connectedSettlementNames.length > 0 && plotIsCityCenter) {
					const plotTooltipConnectionsHeader = document.createElement("div");
					plotTooltipConnectionsHeader.classList.add('text-2xs', 'text-center');
					plotTooltipConnectionsHeader?.style.setProperty('font-weight', 'bold');
					plotTooltipConnectionsHeader.innerHTML = Locale.compose("LOC_PLOT_MOD_TCS_CONNECTED");
					this.container.appendChild(plotTooltipConnectionsHeader);
					
					const plotTooltipConnectionsContainer = document.createElement("div");
					plotTooltipConnectionsContainer.classList.add('plot-tooltip__resource-container');
					plotTooltipConnectionsContainer?.style.setProperty('justify-content', 'space-evenly');
					plotTooltipConnectionsContainer?.style.setProperty('align-content', 'center');
					plotTooltipConnectionsContainer?.style.setProperty('flex-wrap', 'wrap');
					plotTooltipConnectionsContainer?.style.setProperty('padding-right', '0.5rem');
					plotTooltipConnectionsContainer?.style.setProperty('padding-left', '0.5rem');
					
					for (const settlementName of connectedSettlementNames) {
						const plotTooltipConnectionsElement = document.createElement("div");
						plotTooltipConnectionsElement.classList.add('text-2xs', 'text-left');
						plotTooltipConnectionsElement?.style.setProperty('margin-right', '0.5rem');
						plotTooltipConnectionsElement.innerHTML = Locale.stylize(settlementName);
						plotTooltipConnectionsContainer.appendChild(plotTooltipConnectionsElement);
					}
					
					this.container.appendChild(plotTooltipConnectionsContainer);
					
					//plotTooltipConnectionsContainer?.style.setProperty('', '');
				}
				/*if (connectedSettlementString != "" && plotIsCityCenter) {
					this.addHorizontalRule();
					const plotTooltipConnections = document.createElement("div");
					plotTooltipConnections.classList.add('text-2xs', 'text-center');
					const tooltipConnectionsString = Locale.compose("LOC_PLOT_MOD_TCS_CONNECTED") + " " + connectedSettlementString;
					plotTooltipConnections.innerHTML = tooltipConnectionsString;
					this.container.appendChild(plotTooltipConnections);
				}*/
				
				// Trade Routes
				/*if (tradeRoutes) {
					this.addHorizontalRule();
					const plotTooltipTradeRoutes = document.createElement("div");
					plotTooltipTradeRoutes.classList.add('text-2xs', 'text-center');
					plotTooltipTradeRoutes.innerHTML = Locale.compose("LOC_PLOT_MOD_TCS_TRADE_ROUTES", tradeRoutes);
					this.container.appendChild(plotTooltipTradeRoutes);
				}*/
			}
		}
		
		//Hmm...do i need this?
		const plotTooltipConqueror = this.getConquerorInfo(districtId);
		if (plotTooltipConqueror) {
			this.container.appendChild(plotTooltipConqueror);
		}
    }
    getConquerorInfo(districtId) {
        if (!districtId) {
            return null;
        }
        const district = Districts.get(districtId);
        if (!district || !ComponentID.isValid(districtId)) {
            console.error(`plot-tooltip: couldn't find any district with the given id: ${districtId}`);
            return null;
        }
        if (district.owner != district.controllingPlayer) {
            const conqueror = Players.get(district.controllingPlayer);
            if (!conqueror) {
                console.error(`plot-tooltip: couldn't find any civilization with the given player ${district.controllingPlayer}`);
                return null;
            }
            if (conqueror.isIndependent) {
                const plotTooltipOwnerLeader = document.createElement("div");
                plotTooltipOwnerLeader.classList.add("plot-tooltip__owner-leader-text");
                const label = Locale.compose("{1_Term}: {2_Subject}", "LOC_PLOT_TOOLTIP_CONQUEROR", "LOC_PLOT_TOOLTIP_INDEPENDENT_CONQUEROR");
                plotTooltipOwnerLeader.innerHTML = label;
                return plotTooltipOwnerLeader;
            }
            else {
                const conquerorName = Locale.compose(conqueror.civilizationFullName);
                const plotTooltipConqueredCiv = document.createElement("div");
                plotTooltipConqueredCiv.classList.add("plot-tooltip__owner-civ-text");
                const label = Locale.compose("{1_Term}: {2_Subject}", "LOC_PLOT_TOOLTIP_CONQUEROR", conquerorName);
                plotTooltipConqueredCiv.innerHTML = label;
                return plotTooltipConqueredCiv;
            }
        }
        else {
            return null;
        }
    }
    getRiverLabel(location) {
        const riverType = GameplayMap.getRiverType(location.x, location.y);
        if (riverType != RiverTypes.NO_RIVER) {
            let riverNameLabel = GameplayMap.getRiverName(location.x, location.y);
            if (!riverNameLabel) {
                switch (riverType) {
                    case RiverTypes.RIVER_MINOR:
                        riverNameLabel = "LOC_MINOR_RIVER_NAME";
                        break;
                    case RiverTypes.RIVER_NAVIGABLE:
                        riverNameLabel = "LOC_NAVIGABLE_RIVER_NAME";
                        break;
                }
            }
            return riverNameLabel;
        }
        else {
            return "";
        }
    }
    getFeatureLabel(location) {
		// Unlike the other functions, this returns an Object
		// Object.plotIsNaturalWonder
		// Object.featureLabel
		// Object.featureTooltip
        let featureLabel = '';
        const featureType = GameplayMap.getFeatureType(location.x, location.y);
        const feature = GameInfo.Features.lookup(featureType);
		const plotIsNaturalWonder = (feature && GameInfo.Feature_NaturalWonders.find(e => e.FeatureType == feature.FeatureType)) ? true : false;
		const featureTooltip = (plotIsNaturalWonder) ? Locale.compose(feature.Description) : undefined;
		
        if (feature) {
            if (GameplayMap.isVolcano(location.x, location.y)) {
				const active = GameplayMap.isVolcanoActive(location.x, location.y);
				const volcanoStatus = (active) ? 'LOC_VOLCANO_ACTIVE' : 'LOC_PLOT_MOD_TCS_DORMANT';
				const volcanoName = GameplayMap.getVolcanoName(location.x, location.y);
				const volcanoDetailsKey = (volcanoName) ? 'LOC_UI_NAMED_VOLCANO_DETAILS' : 'LOC_UI_VOLCANO_DETAILS';
				featureLabel = (volcanoName) ? Locale.compose("LOC_PLOT_MOD_TCS_VOLCANO", volcanoName, volcanoStatus) : Locale.compose("LOC_PLOT_MOD_TCS_UNNAMED_VOLCANO", volcanoStatus);
			}
			else {
				featureLabel = Locale.compose(feature.Name);
			}
        }
        return {
			plotIsNaturalWonder : plotIsNaturalWonder, 
			featureLabel : featureLabel,
			featureTooltip : featureTooltip
		};
    }
    addUnitInfo(location) {
        const localPlayerID = GameContext.localObserverID;
        if (GameplayMap.getRevealedState(localPlayerID, location.x, location.y) != RevealedStates.VISIBLE) {
            return this;
        }
        let topUnit = this.getTopUnit(location);
        if (topUnit) {
            if (!Visibility.isVisible(localPlayerID, topUnit?.id)) {
                return this;
            }
        }
        else {
            return this;
        }
        const player = Players.get(topUnit.owner);
        if (!player) {
            return this;
        }      
		
		// Build Element: Unit Header
		this.addTitleLine(Locale.compose("LOC_PLOT_MOD_TCS_UNITS"));
		
		// Build Elements: Unit
		let plotUnits = MapUnits.getUnits(location.x, location.y);
		const toolTipHorizontalRule = document.createElement("div");
		toolTipHorizontalRule.classList.add("plot-tooltip__horizontalRule");
		this.container.appendChild(toolTipHorizontalRule);
		const unitContainer = document.createElement("div");
		for (let i = 0; i < plotUnits.length && i < 4; i++) {
			let plotUnit = Units.get(plotUnits[i]);
			let unitName = Locale.compose(plotUnit.name);
			const toolTipUnitInfo = document.createElement("div");
			if (player.id != localPlayerID && i == 0) {
				const playerDiplomacy = player.Diplomacy;
				if (playerDiplomacy.isAtWarWith(GameContext.localPlayerID)) {
					unitContainer.classList.add("plot-tooltip__district-container");
				}
			}
			toolTipUnitInfo.classList.add('text-center',"plot-tooltip__unitInfo");
			if (player.id == localPlayerID) {
				toolTipUnitInfo.innerHTML = Locale.compose("LOC_PLOT_MOD_TCS_TOP_UNIT", unitName, Locale.compose("LOC_PLOT_MOD_TCS_YOURS"));
			}
			else {
				toolTipUnitInfo.innerHTML = Locale.compose("LOC_PLOT_MOD_TCS_TOP_UNIT", unitName, Locale.compose(player.name));
			}
			unitContainer.appendChild(toolTipUnitInfo);
			
		}
		if (plotUnits.length > 3) {
			const toolTipUnitInfo = document.createElement("div");
			toolTipUnitInfo.classList.add('text-center',"plot-tooltip__unitInfo");
			toolTipUnitInfo.innerHTML = Locale.compose("LOC_PLOT_MOD_TCS_ADDITIONAL_UNITS", (plotUnits.length - 3));
			unitContainer.appendChild(toolTipUnitInfo);
		}
		this.container.appendChild(unitContainer);
        return this;
    }
    /**
     * Add to a plot tooltip any yields that are greater than 0 for that plot
     * @param {float2} location The X,Y plot location.
     * @param {playerId} playerID The player associated with the request.
     */
    addPlotYields(location, playerID) {
        const fragment = document.createDocumentFragment();
        let maxValueLength = 0;
        GameInfo.Yields.forEach(yield_define => {
            const yield_amount = GameplayMap.getYield(location.x, location.y, yield_define.YieldType, playerID);
            if (yield_amount > 0) {
                const tooltipIndividualYieldFlex = document.createElement("div");
                tooltipIndividualYieldFlex.classList.add("plot-tooltip__IndividualYieldFlex");
                tooltipIndividualYieldFlex.ariaLabel = `${Locale.toNumber(yield_amount)} ${Locale.compose(yield_define.Name)}`;
                fragment.appendChild(tooltipIndividualYieldFlex);
                const yieldIconCSS = UI.getIconCSS(yield_define.YieldType, "YIELD");
                const yieldIconShadow = document.createElement("div");
                yieldIconShadow.classList.add("plot-tooltip__IndividualYieldIcons-Shadow");
                yieldIconShadow.style.backgroundImage = yieldIconCSS;
                tooltipIndividualYieldFlex.appendChild(yieldIconShadow);
                const yieldIcon = document.createElement("div");
                yieldIcon.classList.add("plot-tooltip__IndividualYieldIcons");
                yieldIcon.style.backgroundImage = yieldIconCSS;
                yieldIconShadow.appendChild(yieldIcon);
                const toolTipIndividualYieldValues = document.createElement("div");
                toolTipIndividualYieldValues.classList.add("plot-tooltip__IndividualYieldValues");
                const value = yield_amount.toString();
                maxValueLength = Math.max(maxValueLength, value.length);
                toolTipIndividualYieldValues.textContent = value;
                tooltipIndividualYieldFlex.appendChild(toolTipIndividualYieldValues);
            }
        });
        this.yieldsFlexbox.appendChild(fragment);
        // Give all the yields extra room if one of them has extra digits, to keep the spacing even.
        this.yieldsFlexbox.classList.remove('resourcesFlex--double-digits', 'resourcesFlex--triple-digits');
        if (maxValueLength > 2) {
            this.yieldsFlexbox.classList.add(maxValueLength > 3 ? 'resourcesFlex--triple-digits' : 'resourcesFlex--double-digits');
        }
    }
}
TooltipManager.registerPlotType('plot', PlotTooltipPriority.LOW, new PlotTooltipType());

//# sourceMappingURL=file:///base-standard/ui/tooltips/plot-tooltip.js.map
